--[[ ---------------------------------------------------------------------------
Name:	SimpleRaidFrameOverlays
Author:	Gallowbraid of Terokkar
About:	A very simple enhancement of the built-in raid frames that allows
        overlay colors to be displayed on raid frames for units that have 
        any sort of debuff. This mimics the functionality of almost all raid
        frame addons without much overhead.
----------------------------------------------------------------------------- ]]

local SRFO = LibStub("AceAddon-3.0"):GetAddon("SimpleRaidFrameOverlays")
local L = LibStub("AceLocale-3.0"):GetLocale("SimpleRaidFrameOverlays")
local H = SRFO:NewModule("Highlight", "AceEvent-3.0")

local backdrop = {
	bgFile = "Interface\\Buttons\\WHITE8X8",
	edgeFile = "Interface\\Buttons\\WHITE8X8",
	tile = false,
	tileSize = 0,
	edgeSize = 2,
	insets = { left=0, right=0, top=0, bottom=0 },
}

SRFO.defaults.profile.highlight = {
	multiplehighlights = 1,
	highlightparty = true,
	highlightpartypets = false,
	highlightraid = true,
	highlightraidpets = false,
	borderpadding = 0,
	borderwidth = 2,
	borderinset = 0,
	separation = 0,
	debuff = {
		onlydispellable = false,
		ignoreunknown = false,
		curse = {
			color = {1.0, 0.0, 1.0, 0.5},
		},
		disease = {
			color = {0.521568, 0.368627, 0.258823, 0.5,},
		},
		magic = {
			color = {0.0, 0.0, 1.0, 0.5,},
		},
		poison = {
			color = {0.0, 1.0, 0.0, 0.5,},
		},
		unknown = {
			color = {1.0, 0.0, 0.0, 0.5},
		},
		filters = {},
	},
	custom = {
		custom1 = {
			filters = {},
			color = {1.0, 0.5, 0.2, 0.5,},
		},
		custom2 = {
			filters = {},
			color = {0.0, 1.0, 1.0, 0.5,},
		},
		custom3 = {
			filters = {},
			color = {1.0, 1.0, 0.0, 0.5,},
		},
	},
	test = {
		curse = false,
		disease = false,
		magic = false,
		poison = false,
		unknown = false,
		custom1 = false,
		custom2 = false,
		custom3 = false,
	},
}

SRFO.options.args.highlight = {
	type = 'group',
	name = L["Highlight"],
	args = {
		enable = {
			type = 'toggle',
			name = L["Enable"],
			desc = L["Enable"],
			get = function(info) return H:IsEnabled() end,
			set = function(info, value) if value then H:Enable() else H:Disable() end end,
			width = "full",
			order = 1,
		},
		multiplehighlights = {
			type = 'toggle',
			name = L["MultipleHighlights"],
			desc = L["MultipleHighlightsDesc"],
			get = function(info) return info.handler.db.profile.highlight.multiplehighlights end,
			set = function(info, value) 
				info.handler.db.profile.highlight.multiplehighlights = value
				SRFO:UnitFramesChanged()
			end,
			width = "full",
			order = 2,
		},
		highlightparty = {
			type = 'toggle',
			name = L["HighlightPartyFrames"],
			desc = L["HighlightPartyFramesDesc"],
			get = function(info) return info.handler.db.profile.highlight.highlightparty end,
			set = function(info, value)
				info.handler.db.profile.highlight.highlightparty = value
				SRFO:UnitFramesChanged()
			end,
			width = "full",
			order = 3,
		},
		highlightpartypets = {
			type = 'toggle',
			name = L["HighlightPartyPetFrames"],
			desc = L["HighlightPartyPetFramesDesc"],
			get = function(info) return info.handler.db.profile.highlight.highlightpartypets end,
			set = function(info, value)
				info.handler.db.profile.highlight.highlightpartypets = value
				SRFO:UnitFramesChanged()
			end,
			width = "full",
			order = 4,
		},
		highlightraid = {
			type = 'toggle',
			name = L["HighlightRaidFrames"],
			desc = L["HighlightRaidFramesDesc"],
			get = function(info) return info.handler.db.profile.highlight.highlightraid end,
			set = function(info, value)
				info.handler.db.profile.highlight.highlightraid = value
				SRFO:UnitFramesChanged()
			end,
			width = "full",
			order = 5,
		},
		highlightraidpets = {
			type = 'toggle',
			name = L["HighlightRaidPetFrames"],
			desc = L["HighlightRaidPetFramesDesc"],
			get = function(info) return info.handler.db.profile.highlight.highlightraidpets end,
			set = function(info, value)
				info.handler.db.profile.highlight.highlightraidpets = value
				SRFO:UnitFramesChanged()
			end,
			width = "full",
			order = 6,
		},
		borderpadding = {
			type = 'range',
			name = L["HighlightBorderPadding"],
			desc = L["HighlightBorderPaddingDesc"],
			min = 0,
			max = 20,
			softMin = 0,
			softMax = 20,
			step = 1,
			bigStep = 1,
			get = function(info) return info.handler.db.profile.highlight.borderpadding end,
			set = function(info, value)
				info.handler.db.profile.highlight.borderpadding = value
				H:UpdateBackdrop()
				SRFO:UnitFramesChanged()
			end,
			width = "full",
			order = 7,
		},
		borderwidth = {
			type = 'range',
			name = L["HighlightBorderWidth"],
			desc = L["HighlightBorderWidthDesc"],
			min = 0,
			max = 20,
			softMin = 0,
			softMax = 20,
			step = 1,
			bigStep = 1,
			get = function(info) return info.handler.db.profile.highlight.borderwidth end,
			set = function(info, value)
				info.handler.db.profile.highlight.borderwidth = value
				H:UpdateBackdrop()
				SRFO:UnitFramesChanged()
			end,
			width = "full",
			order = 8,
		},
		borderinset = {
			type = 'range',
			name = L["HighlightBorderInset"],
			desc = L["HighlightBorderInsetDesc"],
			min = 0,
			max = 20,
			softMin = 0,
			softMax = 20,
			step = 1,
			bigStep = 1,
			get = function(info) return info.handler.db.profile.highlight.borderinset end,
			set = function(info, value)
				info.handler.db.profile.highlight.borderinset = value
				H:UpdateBackdrop()
				SRFO:UnitFramesChanged()
			end,
			width = "full",
			order = 9,
		},
		separation = {
			type = 'range',
			name = L["HighlightSeparation"],
			desc = L["HighlightSeparationDesc"],
			min = 0,
			max = 20,
			softMin = 0,
			softMax = 20,
			step = 1,
			bigStep = 1,
			get = function(info) return info.handler.db.profile.highlight.separation end,
			set = function(info, value)
				info.handler.db.profile.highlight.separation = value
				H:UpdateBackdrop()
				SRFO:UnitFramesChanged()
			end,
			width = "full",
			order = 10,
		},
		debuffs = {
			type = 'group',
			name = L["Debuffs"],
			order = 1,
			args = {
				onlydispellable = {
					type = 'toggle',
					name = L["OnlyDispellable"],
					desc = L["OnlyDispellableDesc"],
					get = function(info) return info.handler.db.profile.highlight.debuff.onlydispellable end,
					set = function(info, value)
						info.handler.db.profile.highlight.debuff.onlydispellable = value
						SRFO:UnitFramesChanged()
					end,
					order = 0,
					width = "full",
				},
				ignoreunknown = {
					type = 'toggle',
					name = L["IgnoreUnknownDebuffs"],
					desc = L["IgnoreUnknownDebuffsDesc"],
					get = function(info) return info.handler.db.profile.highlight.debuff.ignoreunknown end,
					set = function(info, value)
						info.handler.db.profile.highlight.debuff.ignoreunknown = value
						SRFO:UnitFramesChanged()
					end,
					order = 1,
					width = "full",
				},
				cursecolor = {
					type = 'color',
					name = L["CurseHighlightColor"],
					desc = L["CurseHighlightColorDesc"],
					hasAlpha = true,
					get = function(info) return unpack(info.handler.db.profile.highlight.debuff.curse.color) end,
					set = function(info, r, g, b, a)
						local color = info.handler.db.profile.highlight.debuff.curse.color
						color[1], color[2], color[3], color[4] = r, g, b, a
						SRFO:UnitFramesChanged()
					end,
					order = 2,
					width = "full",
				},
				diseasecolor = {
					type = 'color',
					name = L["DiseaseHighlightColor"],
					desc = L["DiseaseHighlightColorDesc"],
					hasAlpha = true,
					get = function(info) return unpack(info.handler.db.profile.highlight.debuff.disease.color) end,
					set = function(info, r, g, b, a)
						local color = info.handler.db.profile.highlight.debuff.disease.color
						color[1], color[2], color[3], color[4] = r, g, b, a
						SRFO:UnitFramesChanged()
					end,
					order = 3,
					width = "full",
				},
				poisoncolor = {
					type = 'color',
					name = L["PoisonHighlightColor"],
					desc = L["PoisonHighlightColorDesc"],
					hasAlpha = true,
					get = function(info) return unpack(info.handler.db.profile.highlight.debuff.poison.color) end,
					set = function(info, r, g, b, a)
						local color = info.handler.db.profile.highlight.debuff.poison.color
						color[1], color[2], color[3], color[4] = r, g, b, a
						SRFO:UnitFramesChanged()
					end,
					order = 4,
					width = "full",
				},
				magiccolor = {
					type = 'color',
					name = L["MagicHighlightColor"],
					desc = L["MagicHighlightColorDesc"],
					hasAlpha = true,
					get = function(info) return unpack(info.handler.db.profile.highlight.debuff.magic.color) end,
					set = function(info, r, g, b, a)
						local color = info.handler.db.profile.highlight.debuff.magic.color
						color[1], color[2], color[3], color[4] = r, g, b, a
						SRFO:UnitFramesChanged()
					end,
					order = 5,
					width = "full",
				},
				unknowncolor = {
					type = 'color',
					name = L["UnknownDebuffColor"],
					desc = L["UnknownDebuffColorDesc"],
					hasAlpha = true,
					get = function(info) return unpack(info.handler.db.profile.highlight.debuff.unknown.color) end,
					set = function(info, r, g, b, a)
						local color = info.handler.db.profile.highlight.debuff.unknown.color
						color[1], color[2], color[3], color[4] = r, g, b, a
						SRFO:UnitFramesChanged()
					end,
					order = 6,
					width = "full",
				},
				filters = {
					type = 'input',
					name = L["DebuffFilters"],
					desc = L["DebuffFiltersDesc"],
					get = function(info)
						return table.concat(info.handler.db.profile.highlight.debuff.filters, "\n")
					end,
					set = function(info, value)
						info.handler.db.profile.highlight.debuff.filters = {strsplit("\n", value)}
						SRFO:UnitFramesChanged()
					end,
					multiline = true,
					order = 7,
					width = "full",
				},
			},
		},
		custom = {
			type = 'group',
			name = L["Custom"],
			order = 2,
			args = {
				customheading1 = {
					type = 'header',
					name = L["CustomHighlightHeading"] .. " 1",
					order = 7,
				},
				customfilter1 = {
					type = 'input',
					name = L["CustomHighlightFilters"],
					desc = L["CustomHighlightFiltersDesc"],
					get = function(info)
						return table.concat(info.handler.db.profile.highlight.custom.custom1.filters)
					end,
					set = function(info, value)
						info.handler.db.profile.highlight.custom.custom1.filters = {strsplit("\n", value)}
						SRFO:UnitFramesChanged()
					end,
					multiline = true,
					width = "full",
					order = 8,
				},
				customcolor1 = {
					type = 'color',
					name = L["CustomHighlightColor"],
					desc = L["CustomHighlightColorDesc"],
					hasAlpha = true,
					get = function(info)
						return unpack(info.handler.db.profile.highlight.custom.custom1.color)
					end,
					set = function(info, r, g, b, a)
						local color = info.handler.db.profile.highlight.custom.custom1.color
						color[1], color[2], color[3], color[4] = r, g, b, a
						SRFO:UnitFramesChanged()
					end,
					order = 9,
				},
				customheading2 = {
					type = 'header',
					name = L["CustomHighlightHeading"] .. " 2",
					order = 10,
				},
				customfilter2 = {
					type = 'input',
					name = L["CustomHighlightFilters"],
					desc = L["CustomHighlightFiltersDesc"],
					get = function(info)
						return table.concat(info.handler.db.profile.highlight.custom.custom2.filters)
					end,
					set = function(info, value)
						info.handler.db.profile.highlight.custom.custom2.filters = {strsplit("\n", value)}
						SRFO:UnitFramesChanged()
					end,
					multiline = true,
					width = "full",
					order = 11,
				},
				customcolor2 = {
					type = 'color',
					name = L["CustomHighlightColor"],
					desc = L["CustomHighlightColorDesc"],
					hasAlpha = true,
					get = function(info)
						return unpack(info.handler.db.profile.highlight.custom.custom2.color)
					end,
					set = function(info, r, g, b, a)
						local color = info.handler.db.profile.highlight.custom.custom2.color
						color[1], color[2], color[3], color[4] = r, g, b, a
						SRFO:UnitFramesChanged()
					end,
					order = 12,
				},
				customheading3 = {
					type = 'header',
					name = L["CustomHighlightHeading"] .. " 3",
					order = 13,
				},
				customfilter3 = {
					type = 'input',
					name = L["CustomHighlightFilters"],
					desc = L["CustomHighlightFiltersDesc"],
					get = function(info)
						return table.concat(info.handler.db.profile.highlight.custom.custom3.filters)
					end,
					set = function(info, value)
						info.handler.db.profile.highlight.custom.custom3.filters = {strsplit("\n", value)}
						SRFO:UnitFramesChanged()
					end,
					multiline = true,
					width = "full",
					order = 14,
				},
				customcolor3 = {
					type = 'color',
					name = L["CustomHighlightColor"],
					desc = L["CustomHighlightColorDesc"],
					hasAlpha = true,
					get = function(info)
						return unpack(info.handler.db.profile.highlight.custom.custom3.color)
					end,
					set = function(info, r, g, b, a)
						local color = info.handler.db.profile.highlight.custom.custom3.color
						color[1], color[2], color[3], color[4] = r, g, b, a
						SRFO:UnitFramesChanged()
					end,
					order = 15,
				},
			},
		},
		test = {
			type = 'group',
			name = L["Testing"],
			order = 3,
			args = {
				curse = {
					type = 'toggle',
					name = L["TestCurseHighlight"],
					desc = L["TestCurseHighlightDesc"],
					get = function(info) return info.handler.db.profile.highlight.test.curse end,
					set = function(info, value)
						info.handler.db.profile.highlight.test.curse = value
						SRFO:UnitFramesChanged()
					end,
					width = "full",
					order = 1,
				},
				disease = {
					type = 'toggle',
					name = L["TestDiseaseHighlight"],
					desc = L["TestDiseaseHighlightDesc"],
					get = function(info) return info.handler.db.profile.highlight.test.disease end,
					set = function(info, value) 
						info.handler.db.profile.highlight.test.disease = value 
						SRFO:UnitFramesChanged()
					end,
					width = "full",
					order = 2,
				},
				magic = {
					type = 'toggle',
					name = L["TestMagicHighlight"],
					desc = L["TestMagicHighlightDesc"],
					get = function(info) return info.handler.db.profile.highlight.test.magic end,
					set = function(info, value)
						info.handler.db.profile.highlight.test.magic = value
						SRFO:UnitFramesChanged()
					end,
					width = "full",
					order = 3,
				},
				poison = {
					type = 'toggle',
					name = L["TestPoisonHighlight"],
					desc = L["TestPoisonHighlightDesc"],
					get = function(info) return info.handler.db.profile.highlight.test.poison end,
					set = function(info, value)
						info.handler.db.profile.highlight.test.poison = value
						SRFO:UnitFramesChanged()
					end,
					width = "full",
					order = 4,
				},
				unknown = {
					type = 'toggle',
					name = L["TestUnknownHighlight"],
					desc = L["TestUnknownHighlightDesc"],
					get = function(info) return info.handler.db.profile.highlight.test.unknown end,
					set = function(info, value)
						info.handler.db.profile.highlight.test.unknown = value
						SRFO:UnitFramesChanged()
					end,
					width = "full",
					order = 5,
				},
				custom1 = {
					type = 'toggle',
					name = L["TestCustomHighlight"] .. " 1",
					desc = L["TestCustomHighlightDesc"],
					get = function(info) return info.handler.db.profile.highlight.test.custom1 end,
					set = function(info, value)
						info.handler.db.profile.highlight.test.custom1 = value
						SRFO:UnitFramesChanged()
					end,
					width = "full",
					order = 6,
				},
				custom2 = {
					type = 'toggle',
					name = L["TestCustomHighlight"] .. " 2",
					desc = L["TestCustomHighlightDesc"],
					get = function(info) return info.handler.db.profile.highlight.test.custom2 end,
					set = function(info, value)
						info.handler.db.profile.highlight.test.custom2 = value
						SRFO:UnitFramesChanged()
					end,
					width = "full",
					order = 7,
				},
				custom3 = {
					type = 'toggle',
					name = L["TestCustomHighlight"] .. " 3",
					desc = L["TestCustomHighlightDesc"],
					get = function(info) return info.handler.db.profile.highlight.test.custom3 end,
					set = function(info, value)
						info.handler.db.profile.highlight.test.custom3 = value
						SRFO:UnitFramesChanged()
					end,
					width = "full",
					order = 8,
				},
			},
		},
	},
}

-- Register the events for this module.
function H:OnEnable()
	self:RegisterEvent("PARTY_MEMBERS_CHANGED", "UpdateAddonStatus")
	if(GetNumPartyMembers() > 0) then
		self:RegisterEvent("UNIT_AURA", "UpdateHighlight")
	end
	if(SRFO:InTestingMode()) then
		self:UpdateForTesting()
	end
end

-- Unregister the events for this module.
function H:OnDisable()
	self:UnregisterEvent("PARTY_MEMBERS_CHANGED")
	self:UnregisterEvent("UNIT_AURA")
	self:CleanFrames()
end

-- If the player has joined a party, start listening for auras.
function H:UpdateAddonStatus()
	if(GetNumPartyMembers() > 0 or GetNumRaidMembers() > 0) then
		self:RegisterEvent("UNIT_AURA", "UpdateHighlight")
		-- Update all of the units with the current auras.
		if(GetNumRaidMembers() > 0) then
			for i=1,GetNumRaidMembers() do
				self:UpdateHighlight("UNIT_AURA", "raid" .. tostring(i))
				self:UpdateHighlight("UNIT_AURA", "raidpet" .. tostring(i))
			end
		elseif(GetNumPartyMembers() > 0) then
			for i=1,GetNumPartyMembers() do
				self:UpdateHighlight("UNIT_AURA", "party" .. tostring(i))
				self:UpdateHighlight("UNIT_AURA", "partypet" .. tostring(i))
			end
			self:UpdateHighlight("UNIT_AURA", "player")
			self:UpdateHighlight("UNIT_AURA", "pet")
		end
	else
		self:UnregisterEvent("UNIT_AURA")
	end
	if(SRFO:InTestingMode()) then
		self:UpdateForTesting()
	end
end

local OverlayData = {}
local OrphanedFrames = {}
local OrphanFather = CreateFrame("Frame")

-- Updates the highlight on the given unit.
-- @param event
--   The event that resulted in this method being called.
-- @param unit
--   The unit whose highlight to update.
function H:UpdateHighlight(event, unit)
	-- Check to see that the unit is one of the valid units we're going to put an overlay on.
	local unitType = 0 -- 1-party, 2-partypet, 3-raid, 4-raidpet
	if(SRFO.db.profile.highlight.highlightparty and SRFO:IsPartyUnit(unit)) then
		unitType = 1
	elseif(SRFO.db.profile.highlight.highlightpartypets and SRFO:IsPartyPetUnit(unit)) then
		unitType = 2
	elseif(SRFO.db.profile.highlight.highlightraid and SRFO:IsRaidUnit(unit)) then
		unitType = 3
	elseif(SRFO.db.profile.highlight.highlightraidpets and SRFO:IsRaidPetUnit(unit)) then
		unitType = 4
	end
	if(unitType > 0) then
		-- Check to see what debuffs the unit has
		local check = 0
		if(SRFO.db.profile.highlight.debuff.onlydispellable) then check = 1 end
		local typeFlags = {}
		-- If we're in testing mode, use the testing flags.
		if(event == "TEST") then
			typeFlags["Curse"] = SRFO.db.profile.highlight.test.curse
			typeFlags["Disease"] = SRFO.db.profile.highlight.test.disease
			typeFlags["Magic"] = SRFO.db.profile.highlight.test.magic
			typeFlags["Poison"] = SRFO.db.profile.highlight.test.poison
			typeFlags["Unknown"] = SRFO.db.profile.highlight.test.unknown
			typeFlags["Custom1"] = SRFO.db.profile.highlight.test.custom1
			typeFlags["Custom2"] = SRFO.db.profile.highlight.test.custom2
			typeFlags["Custom3"] = SRFO.db.profile.highlight.test.custom3
		-- Otherwise, loop through all of the debuffs and analyze each one.
		else
			for i=1,40 do
				local name, rank, icon, count, debuffType = UnitDebuff(unit, i, check)
				if(name and (check == 0 or self:IsDispellable(debuffType))) then
					-- Check to see if it is in a custom filter
					if(H:IsCustom1(name)) then
						typeFlags["Custom1"] = true
					elseif(H:IsCustom2(name)) then
						typeFlags["Custom2"] = true
					elseif(H:IsCustom3(name)) then
						typeFlags["Custom3"] = true
					-- Check to see if it is not a filtered debuff.
					elseif not self:IsDebuffFiltered(name) then
						-- If unkonwn, only highlight if enabled.
						if(not debuffType) then
							if(not SRFO.db.profile.highlight.debuff.ignoreunknown) then 
								typeFlags["Unknown"] = true
							end
						else
							typeFlags[debuffType] = true
						end
					end
				elseif(not name) then
					break
				end
			end
			for i=1,40 do
				local name = UnitBuff(unit, i)
				if(name) then
					-- Check to see if it is in a custom filter
					if(H:IsCustom1(name)) then
						typeFlags["Custom1"] = true
					elseif(H:IsCustom2(name)) then
						typeFlags["Custom2"] = true
					elseif(H:IsCustom3(name)) then
						typeFlags["Custom3"] = true
					end
				else
					break
				end
			end
		end
		-- Build a list of colors for frames that we'll have to create
		local colors = {}
		if(typeFlags["Curse"]) then
			colors[#colors+1] = {unpack(SRFO.db.profile.highlight.debuff.curse.color)}
		end
		if(typeFlags["Disease"]) then
			colors[#colors+1] = {unpack(SRFO.db.profile.highlight.debuff.disease.color)}
		end
		if(typeFlags["Magic"]) then
			colors[#colors+1] = {unpack(SRFO.db.profile.highlight.debuff.magic.color)}
		end
		if(typeFlags["Poison"]) then
			colors[#colors+1] = {unpack(SRFO.db.profile.highlight.debuff.poison.color)}
		end
		if(typeFlags["Unknown"]) then
			colors[#colors+1] = {unpack(SRFO.db.profile.highlight.debuff.unknown.color)}
		end
		if(typeFlags["Custom1"]) then
			colors[#colors+1] = {unpack(SRFO.db.profile.highlight.custom.custom1.color)}
		end
		if(typeFlags["Custom2"]) then
			colors[#colors+1] = {unpack(SRFO.db.profile.highlight.custom.custom2.color)}
		end
		if(typeFlags["Custom3"]) then
			colors[#colors+1] = {unpack(SRFO.db.profile.highlight.custom.custom3.color)}
		end
		
		-- If the setting for multiple colors is disabled and there are multiple colors,
		-- force it to the unknown color.
		if(not SRFO.db.profile.highlight.multiplehighlights and #colors > 1) then
			colors = {}
			colors[1] = {unpack(SRFO.db.profile.highlight.debuff.unknown.color)}
		end
		-- Create more overlay frames until the number of overlay frames for the unit
		-- frame matches the number of debuffs.
		local overlayData = OverlayData[unit]
		local frames = OverlayData.frames
		while(#colors > self:GetNumOverlayFrames(unit)) do
			self:AddOverlayFrame(unit, unitType)
		end
		-- If there are more overlay frames than debuffs on the unit, clear the overlay
		-- frames until the number matches.
		while(#colors < self:GetNumOverlayFrames(unit)) do
			self:RemoveLastOverlayFrame(unit)
		end
		-- If there are actually highlights, handle them.
		if(#colors > 0) then
			local uf = SRFO:GetUnitFrame(unit, unitType)
			if(uf) then
				-- Now that the number of overlay frames matches, resize and position them 
				-- to account for the number of highlights and set the background color.
				local borderPadding = SRFO.db.profile.highlight.borderpadding
				local separation = SRFO.db.profile.highlight.separation
				local width = uf:GetWidth()-2*borderPadding-(#colors-1)*separation
				local ovWidth = width/#colors
				for i=1,#colors do
					local f = self:GetOverlayFrame(unit, i)
					f:SetHeight(uf:GetHeight()-2*borderPadding)
					f:SetWidth(ovWidth)
					if(i==1) then
						f:SetPoint("TOPLEFT", uf, "TOPLEFT", (i-1)*ovWidth+borderPadding, -borderPadding)
					else
						f:SetPoint("TOPLEFT", uf, "TOPLEFT", (i-1)*ovWidth+borderPadding+(i-1)*separation, -borderPadding)
					end				
					f:SetBackdrop(backdrop)
					f:SetBackdropColor(colors[i][1], colors[i][2], colors[i][3], colors[i][4])
					if(SRFO.db.profile.highlight.borderwidth < 1) then
						f:SetBackdropBorderColor(colors[i][1], colors[i][2], colors[i][3], 0.0)
					else
						f:SetBackdropBorderColor(colors[i][1], colors[i][2], colors[i][3], 1.0)
					end				
				end
			end
		end
	end
end

-- Function called by the addon to clean up any current overlays.
-- This function removes all of the highlight frames from unit frames and stores them in
-- the orphaned list for later use.
function H:CleanFrames()
	for k,v in pairs(OverlayData) do
		for k,v in pairs(v.frames) do
			v:SetParent(OrphanFather)
			OrphanedFrames[#OrphanedFrames+1] = v
			v:Hide()
		end
		v.frames = {}
	end
end

-- Function called by either the addon or the module's testing options changing.
function H:UpdateForTesting()
	if(SRFO:InTestingMode()) then
		self:UpdateHighlight("TEST", "testparty")
		self:UpdateHighlight("TEST", "testpartypet")
		self:UpdateHighlight("TEST", "testraid")
		self:UpdateHighlight("TEST", "testraidpet")
	else
		self:CleanFrames()
		self:UpdateAddonStatus()
	end
end

-- Function called by changing the highlight padding and border width.
function H:UpdateBackdrop()
	backdrop.edgeSize = SRFO.db.profile.highlight.borderwidth
	local inset = SRFO.db.profile.highlight.borderinset+SRFO.db.profile.highlight.borderwidth
	backdrop.insets.left = inset
	backdrop.insets.top = inset
	backdrop.insets.right = inset
	backdrop.insets.bottom = inset
end

-- Indicates whether the given debuff is filtered.
-- @param debuff
--   The name of the debuff.
function H:IsDebuffFiltered(debuff)
	for k,v in pairs(SRFO.db.profile.highlight.debuff.filters) do
		if(debuff == v) then return true end
	end
	return false
end

-- Indicates whether the given debuff is filtered for the first custom frame.
-- @param debuff
--   The name of the debuff.
function H:IsCustom1(debuff)
	for k,v in pairs(SRFO.db.profile.highlight.custom.custom1.filters) do
		if(debuff == v) then return true end
	end
	return false
end

-- Indicates whether the given debuff is filtered for the second custom frame.
-- @param debuff
--   The name of the debuff.
function H:IsCustom2(debuff)
	for k,v in pairs(SRFO.db.profile.highlight.custom.custom2.filters) do
		if(debuff == v) then return true end
	end
	return false
end

-- Indicates whether the given debuff is filtered for the third custom frame.
-- @param debuff
--   The name of the debuff.
function H:IsCustom3(debuff)
	for k,v in pairs(SRFO.db.profile.highlight.custom.custom3.filters) do
		if(debuff == v) then return true end
	end
	return false
end

-- Retrieves the number of highlight frames on the given unit.
-- @param unit
--   The unit to retrieve the frame count for.
function H:GetNumOverlayFrames(unit)
	if(OverlayData[unit] and OverlayData[unit].frames) then
		return #(OverlayData[unit].frames)
	end
	return 0
end

-- Adds a highlight frame to the given unit.
-- @param unit
--   The unit to add the highlight frame to.
-- @param unitType
--   The type of the unit (1-party, 2-party pet, 3-raid, 4-raid pet).
function H:AddOverlayFrame(unit, unitType)
	local uf = SRFO:GetUnitFrame(unit, unitType)
	if(not OverlayData[unit]) then
		OverlayData[unit] = {
			frames = {}
		}
	end
	local frames = OverlayData[unit].frames
	if(#OrphanedFrames > 0) then
		local of = OrphanedFrames[#OrphanedFrames]
		of:SetParent(uf)
		frames[#frames+1] = of
		table.remove(OrphanedFrames, #OrphanedFrames)
	else
		frames[#frames+1] = CreateFrame("Frame", "SRFOHighlightFrame" .. tostring(#frames+1), uf)
	end
	frames[#frames]:SetBackdrop(backdrop)
	frames[#frames]:Show()
end

-- Removes the last highlight frame from the given unit.
-- @param unit
--   The unit to remove a highlight from.
function H:RemoveLastOverlayFrame(unit)
	local frames = OverlayData[unit].frames
	frames[#frames]:SetParent(OrphanFather)
	frames[#frames]:Hide()
	OrphanedFrames[#OrphanedFrames+1] = frames[#frames]
	table.remove(frames, #frames)
end

-- Gets the frame associated to a highlight on a unit frame.
-- @param unit
--   The unit to retrieve the highlight frame for.
-- @param frame
--   The index of the frame (starting at 1).
function H:GetOverlayFrame(unit, frame)
	return OverlayData[unit].frames[frame]
end

-- Table of debuffs that can be dispelled by each class and spells.
local DebuffMap = {
	DRUID = {
		Curse = {
			"Remove Curse",
		},
		Poison = {
			"Abolish Poison",
			"Cure Poison",
		},
	},
	MAGE = {
		Curse = {
			"Remove Curse",
		},
	},
	PALADIN = {
		Disease = {
			"Cleanse",
			"Purify",
		},
		Magic = {
			"Cleanse",
		},
		Poison = {
			"Cleanse",
			"Purify",
		},
	},
	PRIEST = {
		Disease = {
			"Abolish Disease",
			"Cure Disease",
		},
		Magic = {
			"Dispel Magic",
			"Mass Dispel",
		},
	},
	SHAMAN = {
		Disease = {
			"Cure Toxins",
			"Cleanse Spirit",
			"Cleansing Totem",
		},
		Poison = {
			"Cure Toxins",
			"Cleanse Spirit",
			"Cleansing Totem",
		},
	},
	WARLOCK = {
		Magic = {
			"Devour Magic",
			"Dispel",
		},
	},
}

-- Checks to see if the given debuff type is dispellable by the player.
-- @param debuffType
--   The type of the debuff to retrieve: "Curse", "Disease", "Magic", or "Poison".
function H:IsDispellable(debuffType)
	local _, class = UnitClass("player")
	-- If there is an entry in the debuff map table for the player's class and debuff
	-- then loop through all of the spells that class has that can remove the debuff.
	-- If the spell is in the spellbook, then they can dispell the debuff.
	if(DebuffMap[class] and debuffType and DebuffMap[class][debuffType]) then
		for k,v in pairs(DebuffMap[class][debuffType]) do
			local skillType, spellId = GetSpellBookItemInfo(v)
			if(skillType) then
				return true
			end
		end
	end
	return false
end
